@@ -1,11 +1,11 @@
 ;;; vundo.el --- Visual undo tree      -*- lexical-binding: t; -*-
 
-;; Copyright (C) 2019-2020 Free Software Foundation, Inc.
+;; Copyright (C) 2019-2023 Free Software Foundation, Inc.
 ;;
 ;; Author: Yuan Fu <casouri@gmail.com>
 ;; Maintainer: Yuan Fu <casouri@gmail.com>
 ;; URL: https://github.com/casouri/vundo
-;; Version: 2.0.0
+;; Version: 2.2.0
 ;; Keywords: undo, text, editing
 ;; Package-Requires: ((emacs "28.1"))
 ;;
@@ -39,6 +39,11 @@
 ;;
 ;;   a   to go back to the last branching point
 ;;   e   to go forward to the end/tip of the branch
+;;   l   to go to the last saved node
+;;
+;;   m   to mark the current node for diff
+;;   u   to unmark the marked node
+;;   d   to show a diff between the marked (or parent) and current nodes
 ;;
 ;;   q   to quit, you can also type C-g
 ;;
@@ -54,7 +59,7 @@
 ;;
 ;; By default, you need to press RET to “commit” your change and if you
 ;; quit with q or C-g, the changes made by vundo are rolled back. You can
-;; set ‘vundo-roll-back-on-quit’ to nil to disable rolling back.
+;; set `vundo-roll-back-on-quit' to nil to disable rolling back.
 ;;
 ;; Note: vundo.el requires Emacs 28.
 ;;
@@ -79,7 +84,7 @@
 ;;
 ;; Your default font needs to contain these Unicode characters, otherwise
 ;; they look terrible and don’t align. You can find a font that covers
-;; these characters (eg, Symbola, Unifont), and set ‘vundo-default’ face
+;; these characters (eg, Symbola, Unifont), and set `vundo-default' face
 ;; to use that font:
 ;;
 ;;     (set-face-attribute 'vundo-default nil :family "Symbola")
@@ -88,9 +93,7 @@
 ;;
 ;; Vundo doesn’t need to be turned on all the time nor replace the undo
 ;; commands like undo-tree does. Vundo displays the tree horizontally,
-;; whereas undo-tree displays a tree vertically. Vundo doesn’t have many
-;; advanced features that undo-tree does (like showing diff), and most
-;; probably will not add those features in the future.
+;; whereas undo-tree displays a tree vertically.
 
 ;;; Developer:
 ;;
@@ -134,21 +137,21 @@
 ;;
 ;; Position-only records
 ;;
-;; We know how undo works: when undoing, ‘primitive-undo’ looks at
-;; each record in ‘pending-undo-list’ and modifies the buffer
+;; We know how undo works: when undoing, `primitive-undo' looks at
+;; each record in `pending-undo-list' and modifies the buffer
 ;; accordingly, and that modification itself pushes new undo records
-;; into ‘buffer-undo-list’. However, not all undo records introduce
-;; modification, if the record is an integer, ‘primitive-undo’ simply
-;; ‘goto’ that position, which introduces no modification to the
-;; buffer and pushes no undo record to ‘buffer-undo-list’. Normally
+;; into `buffer-undo-list'. However, not all undo records introduce
+;; modification, if the record is an integer, `primitive-undo' simply
+;; `goto' that position, which introduces no modification to the
+;; buffer and pushes no undo record to `buffer-undo-list'. Normally
 ;; position records accompany other buffer-modifying records, but if a
 ;; particular record consists of only position records, we have
-;; trouble: after an undo step, ‘buffer-undo-list’ didn’t grow, as far
+;; trouble: after an undo step, `buffer-undo-list' didn’t grow, as far
 ;; as vundo tree-folding algorithm is concerned, we didn’t move.
-;; Assertions expecting to see new undo records in ‘buffer-undo-list’
+;; Assertions expecting to see new undo records in `buffer-undo-list'
 ;; are also violated. To avoid all these complications, we ignore
 ;; position-only records when generating mod-list in
-;; ‘vundo--mod-list-from’. These records are not removed, but they
+;; `vundo--mod-list-from'. These records are not removed, but they
 ;; can’t harm us now.
 
 ;;; Code:
@@ -156,6 +159,7 @@
 (require 'pcase)
 (require 'cl-lib)
 (require 'seq)
+(require 'subr-x)
 
 ;;; Customization
 
@@ -172,6 +176,10 @@
 (defface vundo-stem '((t . (:inherit vundo-default)))
   "Face for stems between nodes in the undo tree.")
 
+(defface vundo-branch-stem
+  '((t (:inherit vundo-stem :weight bold)))
+  "Face for branching stems in the undo tree.")
+
 (defface vundo-highlight
   '((((background light)) .
      (:inherit vundo-node :weight bold :foreground "red"))
@@ -179,10 +187,27 @@
      (:inherit vundo-node :weight bold :foreground "yellow")))
   "Face for the highlighted node in the undo tree.")
 
+(defface vundo-saved
+  '((((background light)) .
+     (:inherit vundo-node :foreground "dark green"))
+    (((background dark)) .
+     (:inherit vundo-node  :foreground "light green")))
+  "Face for saved nodes in the undo tree.")
+
+(defface vundo-last-saved
+  '((t (:inherit vundo-saved :weight bold)))
+  "Face for the last saved node in the undo tree.")
+
 (defcustom vundo-roll-back-on-quit t
   "If non-nil, vundo will roll back the change when it quits."
   :type 'boolean)
 
+(defcustom vundo-highlight-saved-nodes t
+  "If non-nil, vundo will highlight nodes which have been saved and then modified.
+The face `vundo-saved' is used for saved nodes, except for the
+most recent such node, which receives the face `vundo-last-saved'."
+  :type 'boolean)
+
 (defcustom vundo-window-max-height 3
   "The maximum height of the vundo window."
   :type 'integer)
@@ -246,12 +271,12 @@ By default, the tree is drawn with ASCII characters like this:
     |--o
     \\=`--o
 
-Set this variable to ‘vundo-unicode-symbols’ to use Unicode
+Set this variable to `vundo-unicode-symbols' to use Unicode
 characters."
   :type `(alist :tag "Translation alist"
-		        :key-type (symbol :tag "Part of tree")
-		        :value-type (character :tag "Draw using")
-		        :options ,(mapcar #'car vundo-unicode-symbols)))
+                :key-type (symbol :tag "Part of tree")
+                :value-type (character :tag "Draw using")
+                :options ,(mapcar #'car vundo-unicode-symbols)))
 
 (defcustom vundo-pre-enter-hook nil
   "List of functions to call when entering vundo.
@@ -260,9 +285,9 @@ the user invoked ‘vundo’, before every setup ‘vundo’ does."
   :type 'hook)
 
 (defcustom vundo-post-exit-hook nil
-  "List of functions to call when entering vundo.
+  "List of functions to call when exiting vundo.
 This hook runs in the original buffer the user invoked ‘vundo’,
-after every clean up the exiting function does. Ie, it is the
+after all the clean up the exiting function does. Ie, it is the
 very last thing that happens when vundo exists."
   :type 'hook)
 
@@ -300,13 +325,21 @@ modification."
   (point
    nil
    :type integer
-   :documentation "Marks the text node in the vundo buffer if drawn."))
+   :documentation "Marks the text node in the vundo buffer if drawn.")
+  (timestamp
+   nil
+   :type timestamp
+   :documentation
+   "Timestamp at which this mod altered a saved buffer state.
+If this field is non-nil, the mod contains a timestamp entry in
+the undo list, meaning the previous state was saved to file. This
+field records that timestamp."))
 
 (defun vundo--position-only-p (undo-list)
   "Check if the records at the start of UNDO-LIST are position-only.
 Position-only means all records until to the next undo
 boundary are position records. Position record is just an
-integer (see ‘buffer-undo-list’). Assumes the first element
+integer (see `buffer-undo-list'). Assumes the first element
 of UNDO-LIST is not nil."
   (let ((pos-only t))
     (while (car undo-list)
@@ -330,18 +363,27 @@ If MOD-LIST non-nil, extend on MOD-LIST."
       ;; It's possible the index was exceeded stepping over nil.
       (when (or (null n) (< uidx n))
         ;; Add modification.
-        (unless (vundo--position-only-p undo-list)
-          ;; If this record is position-only, we skip it and don’t add a
-          ;; mod for it. Effectively taking it out of the undo tree.
-          ;; Read ‘Position-only records’ section in Commentary for more
-          ;; explanation.
-          (cl-assert (not (null (car undo-list))))
-          (push (make-vundo-m :undo-list undo-list)
-                new-mlist))
-        ;; Skip through the content of this modification.
-        (while (car undo-list)
-          (setq undo-list (cdr undo-list))
-          (cl-incf uidx))))
+        (let ((pos-only (vundo--position-only-p undo-list))
+              (mod-timestamp nil))
+          (unless pos-only
+            ;; If this record is position-only, we skip it and don’t
+            ;; add a mod for it. Effectively taking it out of the undo
+            ;; tree. Read ‘Position-only records’ section in
+            ;; Commentary for more explanation.
+            (cl-assert (not (null (car undo-list))))
+            (push (make-vundo-m :undo-list undo-list)
+                  new-mlist))
+          ;; Skip through the content of this modification.
+          (while (car undo-list)
+            ;; Is this entry a timestamp?
+            (when (and (consp (car undo-list)) (eq (caar undo-list) t))
+              (setq mod-timestamp (cdar undo-list)))
+            (setq undo-list (cdr undo-list))
+            (cl-incf uidx))
+          ;; If this modification contains a timestamp, the previous
+          ;; state is saved to file.
+          (when (and mod-timestamp (not pos-only))
+            (setf (vundo-m-timestamp (car new-mlist)) mod-timestamp)))))
     ;; Convert to vector.
     (vconcat mod-list new-mlist)))
 
@@ -474,6 +516,36 @@ If FROM non-nil, build from FORM-th modification in MOD-LIST."
                        (vundo--sort-mod (cons mod children)
                                         'reverse))))))))))
 
+;;; Timestamps
+;; buffer-undo-list contains "timestamp entries" like (t . TIMESTAMP)
+;; which capture the file modification time of the saved file which
+;; an undo changed.  During tree draw, we collect the last of these, and
+;; indicated nodes which had been saved specially.
+
+(defvar vundo--last-saved-idx)
+(defvar vundo--orig-buffer)
+
+(defun vundo--mod-timestamp (mod-list idx)
+  "Return a timestamp if the mod in MOD-LIST at IDX has a timestamp."
+  ;; If the next mod’s timestamp is non-nil, this mod/node
+  ;; represents a saved state.
+  (let* ((next-mod-idx (1+ idx))
+         (next-mod (when (< next-mod-idx (length mod-list))
+                     (aref mod-list next-mod-idx))))
+    (and next-mod (vundo-m-timestamp next-mod))))
+
+(defun vundo--node-timestamp (mod-list node)
+  "Return a timestamp from MOD-LIST for NODE, if any.
+In addition to undo-based timestamps, this includes the modtime of the
+current buffer (if unmodified)."
+  (let* ((idx (vundo-m-idx node))
+         (current (vundo--current-node mod-list)))
+    (or (vundo--mod-timestamp mod-list idx)
+        (and (eq node current) (eq idx vundo--last-saved-idx)
+             (with-current-buffer vundo--orig-buffer
+               (and (buffer-file-name)
+                    (not (buffer-modified-p))
+                    (visited-file-modtime)))))))
 ;;; Draw tree
 
 (defun vundo--put-node-at-point (node)
@@ -500,7 +572,7 @@ If FROM non-nil, build from FORM-th modification in MOD-LIST."
 
 (defun vundo--translate (text)
   "Translate each character in TEXT and return translated TEXT.
-Translate according to ‘vundo-glyph-alist’."
+Translate according to `vundo-glyph-alist'."
   (seq-mapcat (lambda (ch)
                 (char-to-string
                  (alist-get
@@ -514,38 +586,61 @@ Translate according to ‘vundo-glyph-alist’."
                   vundo-glyph-alist)))
               text 'string))
 
-(defun vundo--draw-tree (mod-list)
-  "Draw the tree in MOD-LIST in current buffer."
+(defvar vundo--last-saved-idx)
+
+(defun vundo--draw-tree (mod-list orig-buffer-modified)
+  "Draw the tree in MOD-LIST in current buffer.
+ORIG-BUFFER-MODIFIED is t if the original buffer is not saved to
+the disk.  Set `vundo--last-saved-idx' by side-effect,
+corresponding to the index of the last saved node."
   (let* ((root (aref mod-list 0))
          (node-queue (list root))
          (inhibit-read-only t)
-         (inhibit-modification-hooks t))
+         (inhibit-modification-hooks t)
+         (last-saved-idx -1))
     (erase-buffer)
+    (setq vundo--last-saved-idx -1)
     (while node-queue
       (let* ((node (pop node-queue))
              (children (vundo-m-children node))
              (parent (vundo-m-parent node))
-             ;; Is NODE the last child of PARENT?
-             (node-last-child-p
-              (if parent
-                  (eq node (car (last (vundo-m-children parent)))))))
+             (siblings (and parent (vundo-m-children parent)))
+             (only-child-p (and parent (eq (length siblings) 1)))
+             (node-last-child-p (and parent (eq node (car (last siblings)))))
+             (node-idx (vundo-m-idx node))
+             (mod-ts (vundo--mod-timestamp mod-list node-idx))
+             (saved-p (and vundo-highlight-saved-nodes mod-ts))
+             (node-face (if saved-p 'vundo-saved 'vundo-node))
+             (stem-face (if only-child-p 'vundo-stem 'vundo-branch-stem)))
+        (when (and mod-ts (> node-idx last-saved-idx))
+          (setq last-saved-idx node-idx))
         ;; Go to parent.
         (if parent (goto-char (vundo-m-point parent)))
-        (let ((col (max 0 (1- (current-column)))))
+        (let ((room-for-another-rx
+               (rx-to-string
+                `(or (>= ,(if vundo-compact-display 3 4) ?\s) eol))))
           (if (null parent)
               (insert (propertize (vundo--translate "○")
-                                  'face 'vundo-node))
+                                  'face node-face))
             (let ((planned-point (point)))
               ;; If a node is blocking, try next line.
               ;; Example: 1--2--3  Here we want to add a
               ;;             |     child to 1 but is blocked
               ;;             +--4  by that plus sign.
-              (while (not (looking-at (rx (or "    " eol))))
-                (vundo--next-line-at-column col)
-                (unless (looking-at "$")
-                  (delete-char 1))
-                (insert (propertize (vundo--translate "│")
-                                    'face 'vundo-stem)))
+              (while (not (looking-at room-for-another-rx))
+                (vundo--next-line-at-column (max 0 (1- (current-column))))
+                ;; When we go down, we could encounter space, EOL, │,
+                ;; ├, or └. Space and EOL should be replaced by │, ├
+                ;; and └ should be replaced by ├.
+                (let ((replace-char
+                       (if (looking-at
+                            (rx-to-string
+                             `(or ,(vundo--translate "├")
+                                  ,(vundo--translate "└"))))
+                           (vundo--translate "├")
+                         (vundo--translate "│"))))
+                  (unless (eolp) (delete-char 1))
+                  (insert (propertize replace-char 'face stem-face))))
               ;; Make room for inserting the new node.
               (unless (looking-at "$")
                 (delete-char (if vundo-compact-display 2 3)))
@@ -554,25 +649,38 @@ Translate according to ‘vundo-glyph-alist’."
                   (insert (propertize
                            (vundo--translate
                             (if vundo-compact-display "─" "──"))
-                           'face 'vundo-stem)
+                           'face stem-face)
                           (propertize (vundo--translate "○")
-                                      'face 'vundo-node))
-                ;; Delete the previously inserted |.
+                                      'face node-face))
+                ;; We must break the line.  Delete the previously inserted char.
                 (delete-char -1)
                 (insert (propertize
                          (vundo--translate
                           (if node-last-child-p
                               (if vundo-compact-display "└─" "└──")
                             (if vundo-compact-display "├─" "├──")))
-                         'face 'vundo-stem))
+                         'face stem-face))
                 (insert (propertize (vundo--translate "○")
-                                    'face 'vundo-node))))))
+                                    'face node-face))))))
         ;; Store point so we can later come back to this node.
         (setf (vundo-m-point node) (point))
         ;; Associate the text node in buffer with the node object.
         (vundo--put-node-at-point node)
         ;; Depth-first search.
-        (setq node-queue (append children node-queue))))))
+        (setq node-queue (append children node-queue))))
+
+    ;; If the associated buffer is unmodified, the last node must be
+    ;; the last saved node even though it doesn’t (yet) have a next
+    ;; node with a timestamp to indicate that.
+    (setq vundo--last-saved-idx
+          (if orig-buffer-modified
+              (if (> last-saved-idx 0) last-saved-idx nil)
+            (vundo-m-idx (vundo--current-node mod-list))))
+    ;; Update the face of the last saved node (if any).
+    (when (and vundo-highlight-saved-nodes
+               vundo--last-saved-idx)
+      (vundo--highlight-last-saved-node
+       (aref mod-list vundo--last-saved-idx)))))
 
 ;;; Vundo buffer and invocation
 
@@ -588,6 +696,9 @@ WINDOW is the window that was/is displaying the vundo buffer."
       (with-selected-window window
         (kill-buffer-and-window))))
 
+(declare-function vundo-diff "vundo-diff")
+(declare-function vundo-diff-mark "vundo-diff")
+(declare-function vundo-diff-unmark "vundo-diff")
 (defvar vundo-mode-map
   (let ((map (make-sparse-keymap)))
     (define-key map (kbd "f") #'vundo-forward)
@@ -600,20 +711,26 @@ WINDOW is the window that was/is displaying the vundo buffer."
     (define-key map (kbd "<up>") #'vundo-previous)
     (define-key map (kbd "a") #'vundo-stem-root)
     (define-key map (kbd "e") #'vundo-stem-end)
+    (define-key map (kbd "l") #'vundo-goto-last-saved)
     (define-key map (kbd "q") #'vundo-quit)
     (define-key map (kbd "C-g") #'vundo-quit)
     (define-key map (kbd "RET") #'vundo-confirm)
+    (define-key map (kbd "m") #'vundo-diff-mark)
+    (define-key map (kbd "u") #'vundo-diff-unmark)
+    (define-key map (kbd "d") #'vundo-diff)
     (define-key map (kbd "i") #'vundo--inspect)
-    (define-key map (kbd "d") #'vundo--debug)
+    (define-key map (kbd "D") #'vundo--debug)
+
+    (define-key map [remap save-buffer] #'vundo-save)
     map)
-  "Keymap for ‘vundo-mode’.")
+  "Keymap for `vundo-mode'.")
 
 (define-derived-mode vundo-mode special-mode
   "Vundo" "Mode for displaying the undo tree."
   (setq mode-line-format nil
         truncate-lines t
         cursor-type nil)
-  (jit-lock-mode -1)
+  (jit-lock-mode nil)
   (face-remap-add-relative 'default 'vundo-default)
 
   ;; Disable evil-mode, as normal-mode
@@ -623,9 +740,9 @@ WINDOW is the window that was/is displaying the vundo buffer."
     (push 'vundo-mode evil-emacs-state-modes)))
 
 (defvar-local vundo--prev-mod-list nil
-  "Modification list generated by ‘vundo--mod-list-from’.")
+  "Modification list generated by `vundo--mod-list-from'.")
 (defvar-local vundo--prev-mod-hash nil
-  "Modification hash table generated by ‘vundo--update-mapping’.")
+  "Modification hash table generated by `vundo--update-mapping'.")
 (defvar-local vundo--prev-undo-list nil
   "Original buffer's `buffer-undo-list'.")
 (defvar-local vundo--orig-buffer nil
@@ -636,6 +753,12 @@ WINDOW is the window that was/is displaying the vundo buffer."
   "Vundo will roll back to this node.")
 (defvar-local vundo--highlight-overlay nil
   "Overlay used to highlight the selected node.")
+(defvar-local vundo--last-saved-idx nil
+  "The last node index with a timestamp seen.
+This is set by ‘vundo--draw-tree’ and ‘vundo-save’, and used by
+‘vundo-goto-last-saved’ and ‘vundo--highlight-last-saved-node’.")
+(defvar-local vundo--highlight-last-saved-overlay nil
+  "Overlay used to highlight the last saved node.")
 
 (defun vundo--mod-list-trim (mod-list n)
   "Remove MODS from MOD-LIST.
@@ -659,9 +782,9 @@ INCREMENTAL is only applicable when entries are either added or
 removed from undo-list. On the other hand, if some entries are
 removed and some added, do not use INCREMENTAL.
 
-This function modifies ‘vundo--prev-mod-list’,
-‘vundo--prev-mod-hash’, ‘vundo--prev-undo-list’,
-‘vundo--orig-buffer’."
+This function modifies `vundo--prev-mod-list',
+`vundo--prev-mod-hash', `vundo--prev-undo-list',
+`vundo--orig-buffer'."
   (with-current-buffer vundo-buffer
     ;; 1. Setting these to nil makes `vundo--mod-list-from',
     ;; `vundo--update-mapping' and `vundo--build-tree' starts from
@@ -669,7 +792,8 @@ This function modifies ‘vundo--prev-mod-list’,
     (when (not incremental)
       (setq vundo--prev-undo-list nil
             vundo--prev-mod-list nil
-            vundo--prev-mod-hash nil)
+            vundo--prev-mod-hash nil
+            vundo--last-saved-idx nil)
       ;; Give the garbage collector a chance to release
       ;; `buffer-undo-list': GC cannot release cons cells when all
       ;; these stuff are referring to it.
@@ -710,12 +834,14 @@ This function modifies ‘vundo--prev-mod-list’,
                              (length vundo--prev-mod-list))))
       ;; 3. Render buffer. We don't need to redraw the tree if there
       ;; is no change to the nodes.
-      (unless (eq (vundo--latest-buffer-state mod-list)
-                  latest-state)
-        (vundo--draw-tree mod-list))
+      (unless (eq (vundo--latest-buffer-state mod-list) latest-state)
+        (vundo--draw-tree mod-list (with-current-buffer orig-buffer
+                                     (buffer-modified-p))))
+
       ;; Highlight current node.
       (vundo--highlight-node (vundo--current-node mod-list))
       (goto-char (vundo-m-point (vundo--current-node mod-list)))
+
       ;; Update cache.
       (setq vundo--prev-mod-list mod-list
             vundo--prev-mod-hash mod-hash
@@ -734,17 +860,32 @@ This function modifies ‘vundo--prev-mod-list’,
     (overlay-put vundo--highlight-overlay
                  'display (vundo--translate "●"))
     (overlay-put vundo--highlight-overlay
-                 'face 'vundo-highlight))
+                 'face 'vundo-highlight)
+    ;; Make current node’s highlight override last saved node’s
+    ;; highlight, should they collide.
+    (overlay-put vundo--highlight-overlay 'priority 2))
   (move-overlay vundo--highlight-overlay
                 (1- (vundo-m-point node))
                 (vundo-m-point node)))
 
+(defun vundo--highlight-last-saved-node (node)
+  "Highlight NODE as the last saved.
+This moves the overlay `vundo--highlight-last-saved-overlay'."
+  (let ((node-pt (vundo-m-point node)))
+    (unless vundo--highlight-last-saved-overlay
+      (setq vundo--highlight-last-saved-overlay
+            (make-overlay (1- node-pt) node-pt))
+      (overlay-put vundo--highlight-last-saved-overlay 'face 'vundo-last-saved))
+    (move-overlay vundo--highlight-last-saved-overlay (1- node-pt) node-pt)))
+
 ;;;###autoload
 (defun vundo ()
   "Display visual undo for the current buffer."
   (interactive)
   (when (not (consp buffer-undo-list))
     (user-error "There is no undo history"))
+  (when buffer-read-only
+    (user-error "Buffer is read-only"))
   (run-hooks 'vundo-pre-enter-hook)
   (let ((vundo-buf (vundo-1 (current-buffer))))
     (select-window
@@ -819,8 +960,11 @@ Roll back changes if `vundo-roll-back-on-quit' is non-nil."
       vundo--orig-buffer vundo--prev-mod-list))
    (with-current-buffer vundo--orig-buffer
      (setq-local buffer-read-only nil))
-   (let ((orig-buffer vundo--orig-buffer))
+   (let* ((orig-buffer vundo--orig-buffer)
+          (orig-window (get-buffer-window orig-buffer)))
      (kill-buffer-and-window)
+     (when (window-live-p orig-window)
+       (select-window orig-window))
      (with-current-buffer orig-buffer
        (run-hooks 'vundo-post-exit-hook)))))
 
@@ -829,8 +973,11 @@ Roll back changes if `vundo-roll-back-on-quit' is non-nil."
   (interactive)
   (with-current-buffer vundo--orig-buffer
     (setq-local buffer-read-only nil))
-  (let ((orig-buffer vundo--orig-buffer))
+  (let* ((orig-buffer vundo--orig-buffer)
+         (orig-window (get-buffer-window orig-buffer)))
     (kill-buffer-and-window)
+    (when (window-live-p orig-window)
+      (select-window orig-window))
     (with-current-buffer orig-buffer
       (run-hooks 'vundo-post-exit-hook))))
 
@@ -840,7 +987,7 @@ Roll back changes if `vundo-roll-back-on-quit' is non-nil."
   "Calculate the shortest route from FROM to TO node.
 Return (SOURCE STOP1 STOP2 ... DEST), meaning you should undo the
 modifications from DEST to SOURCE. Each STOP is an intermediate
-stop. Eg, (6 5 4 3). Return nil if no valid route."
+stop. Eg, (6 5 4 3). Return nil if there’s no valid route."
   (let (route-list)
     ;; Find all valid routes.
     (dolist (source (vundo--eqv-list-of from))
@@ -866,7 +1013,7 @@ stop. Eg, (6 5 4 3). Return nil if no valid route."
 (defun vundo--list-subtract (l1 l2)
   "Return L1 - L2.
 
-\(vundo--list-subtract '(4 3 2 1) '(2 1))
+\(vundo--list-subtract \='(4 3 2 1) \='(2 1))
 => (4 3)"
   (let ((len1 (length l1))
         (len2 (length l2)))
@@ -895,8 +1042,10 @@ Basically, return the latest non-undo modification in MOD-LIST."
 (defun vundo--move-to-node (current dest orig-buffer mod-list)
   "Move from CURRENT node to DEST node by undoing in ORIG-BUFFER.
 ORIG-BUFFER must be at CURRENT state. MOD-LIST is the list you
-get from ‘vundo--mod-list-from’. You should refresh vundo buffer
-after calling this function.
+get from `vundo--mod-list-from'. You should refresh vundo buffer
+after calling this function. (You can call this function
+repeatedly before refreshing, but moving back-and-forth might not
+work, see docstring of ‘vundo--trim-undo-list’.)
 
 This function modifies the content of ORIG-BUFFER."
   (cl-assert (not (eq current dest)))
@@ -918,7 +1067,7 @@ This function modifies the content of ORIG-BUFFER."
         (with-current-buffer orig-buffer
           (setq-local buffer-read-only t)
           ;; 2. Undo. This will undo modifications in PLANNED-UNDO and
-          ;; add new entries to ‘buffer-undo-list’.
+          ;; add new entries to `buffer-undo-list'.
           (let ((undo-in-progress t))
             (cl-loop
              for step = (- source-idx dest-idx)
@@ -955,29 +1104,37 @@ This function modifies the content of ORIG-BUFFER."
     (error "No possible route")))
 
 (defun vundo--trim-undo-list (buffer current mod-list)
-  "Trim ‘buffer-undo-list’ in BUFFER according to CURRENT and MOD-LIST.
+  "Trim `buffer-undo-list' in BUFFER according to CURRENT and MOD-LIST.
 CURRENT is the current mod, MOD-LIST is the current mod-list.
 
-This function modifies ‘buffer-undo-list’ of BUFFER.
+This function modifies `buffer-undo-list' of BUFFER.
 
-IMPORTANT Relationship between ‘vundo--move-to-node’,
-‘vundo--refresh-buffer’, ‘vundo--trim-undo-list’:
+IMPORTANT Relationship between `vundo--move-to-node',
+`vundo--refresh-buffer', `vundo--trim-undo-list':
 
 Each vundo command cycle roughly works like this:
-1. ‘vundo--refresh-buffer’: ‘buffer-undo-list’ -> mod-list
-2. ‘vundo--move-to-node’: read mod-list, modify ‘buffer-undo-list’
-3. ‘vundo--trim-undo-list’: trim ‘buffer-undo-list’
-1. ‘vundo--refresh-buffer’: ‘buffer-undo-list’ -> mod-list
+1. `vundo--refresh-buffer': `buffer-undo-list' -> mod-list
+2. `vundo--move-to-node': read mod-list, modify `buffer-undo-list'
+3. `vundo--trim-undo-list': trim `buffer-undo-list'
+1. `vundo--refresh-buffer': `buffer-undo-list' -> mod-list
 ...
 
-We can call ‘vundo--move-to-node’ multiple times between two
-‘vundo--refresh-buffer’. But we should only call
-‘vundo--trim-undo-list’ once between two ‘vundo--refresh-buffer’.
-Because if we only trim once, ‘buffer-undo-list’ either shrinks
+We can call `vundo--move-to-node' multiple times between two
+`vundo--refresh-buffer'. But we should only call
+`vundo--trim-undo-list' once between two `vundo--refresh-buffer'.
+Because if we only trim once, `buffer-undo-list' either shrinks
 or expands. But if we trim multiple times after multiple
 movements, it could happen that the undo-list first
 shrinks (trimmed) then expands. In that situation we cannot use
-the INCREMENTAL option in ‘vundo--refresh-buffer’ anymore."
+the INCREMENTAL option in `vundo--refresh-buffer' anymore.
+
+Also, if you move back-end-forth with ‘vundo--move-to-node’, it
+might not work: Suppose undo list is [1 2 3], mod-list is [1 2
+3], now we move back to 2, undo list becomes [1 2 3 2’], but
+before we refresh vundo buffer, mod-list will remain [1 2 3], so
+there’s no route from 2 to 3 (you can only move back). Once
+we refresh the buffer and mod-list is updated to [1 2 3 2’], we
+have a route from 3 to 2 (2’->3)."
   (let ((latest-buffer-state-idx
          ;; Among all the MODs that represents a unique buffer
          ;; state, we find the latest one. Because any node
@@ -999,6 +1156,11 @@ the INCREMENTAL option in ‘vundo--refresh-buffer’ anymore."
         (message "Trimmed to: %s"
                  (vundo-m-idx possible-trim-point))))))
 
+(defvar vundo-after-undo-functions nil
+  "Special hook that runs after `vundo' motions.
+Functions aasigned to this hook are called with one argument: the
+original buffer `vundo' operates on.")
+
 (defun vundo-forward (arg)
   "Move forward ARG nodes in the undo tree.
 If ARG < 0, move backward."
@@ -1027,7 +1189,8 @@ If ARG < 0, move backward."
           vundo--orig-buffer dest vundo--prev-mod-list)
          ;; Refresh display.
          (vundo--refresh-buffer
-          vundo--orig-buffer (current-buffer) 'incremental))))))
+          vundo--orig-buffer (current-buffer) 'incremental))))
+   (run-hook-with-args 'vundo-after-undo-functions vundo--orig-buffer)))
 
 (defun vundo-backward (arg)
   "Move back ARG nodes in the undo tree.
@@ -1057,7 +1220,8 @@ If ARG < 0, move forward."
             vundo--orig-buffer dest vundo--prev-mod-list)
            (vundo--refresh-buffer
             vundo--orig-buffer (current-buffer)
-            'incremental)))))))
+            'incremental)))))
+   (run-hook-with-args 'vundo-after-undo-functions vundo--orig-buffer)))
 
 (defun vundo-previous (arg)
   "Move to node above the current one. Move ARG steps."
@@ -1121,6 +1285,34 @@ If ARG < 0, move forward."
       vundo--orig-buffer (current-buffer)
       'incremental))))
 
+(defun vundo-goto-last-saved ()
+  "Goto the last saved node, if any."
+  (interactive)
+  (when (and vundo--last-saved-idx (>= vundo--last-saved-idx 0))
+    (vundo--check-for-command
+     (when-let* ((this (vundo--current-node vundo--prev-mod-list))
+                 (dest (aref vundo--prev-mod-list vundo--last-saved-idx)))
+       (unless (eq this dest)
+         (vundo--move-to-node
+          this dest vundo--orig-buffer vundo--prev-mod-list)
+         (vundo--trim-undo-list
+          vundo--orig-buffer dest vundo--prev-mod-list)
+         (vundo--refresh-buffer
+          vundo--orig-buffer (current-buffer)
+          'incremental))))))
+
+(defun vundo-save (arg)
+  "Run `save-buffer' with the current buffer Vundo is operating on.
+Accepts the same interactive arfument ARG as ‘save-buffer’."
+  (interactive "p")
+  (vundo--check-for-command
+   (with-current-buffer vundo--orig-buffer
+     (save-buffer arg)))
+  (let* ((cur-node (vundo--current-node vundo--prev-mod-list)))
+    (setq vundo--last-saved-idx (vundo-m-idx cur-node))
+    (when vundo-highlight-saved-nodes
+      (vundo--highlight-last-saved-node cur-node))))
+
 ;;; Debug
 
 (defun vundo--setup-test-buffer ()
@@ -1136,12 +1328,16 @@ TYPE is the type of buffer you want."
   "Print some useful info about the node at point."
   (interactive)
   (let ((node (vundo--get-node-at-point)))
-    (message "Parent: %s States: %s Children: %s"
+    (message "Parent: %s States: %s Children: %s%s"
              (and (vundo-m-parent node)
                   (vundo-m-idx (vundo-m-parent node)))
              (mapcar #'vundo-m-idx (vundo--eqv-list-of node))
              (and (vundo-m-children node)
-                  (mapcar #'vundo-m-idx (vundo-m-children node))))))
+                  (mapcar #'vundo-m-idx (vundo-m-children node)))
+             (if-let* ((ts (vundo--node-timestamp vundo--prev-mod-list node))
+                       ((consp ts)))
+                 (format " Saved: %s" (format-time-string "%F %r" ts))
+               ""))))
 
 (defun vundo--debug ()
   "Make cursor visible and show debug information on movement."
@@ -1163,3 +1359,7 @@ TYPE is the type of buffer you want."
 (provide 'vundo)
 
 ;;; vundo.el ends here
+
+;; Local Variables:
+;; indent-tabs-mode: nil
+;; End:
